// debug option
var LOGGING = false;

// Frame expanding instance
var expander = null;
var buttonlisteners = new Array();
var autoCollapse = false;
var expansionReserved = false;

/* 
   parameters imported from HTML, 
   use <parameter> elements
*/
var importedParameters = null;

// font size
var baseFontSizeDesc    = 10;
var baseFontSizeChannel = 10;
var baseFontUnitDesc    = "px";
var baseFontUnitChannel = "px";
var fontSizeDesc = 10;
var fontSizeChannel = 10;

// TV area 
var channels        = null;
var tableStartTime  = 0;
var ProgramArea     = "tokyo";
var ProgramWave     = "vhf";
var offsetThreshold = 240;
var dayOfData       = new Date();
var dateLine        = 240;

// table parameters 
var displaySpan     = 180;
var offsetMinutes   = 60;

// layout 
var channelFormat     = "";
var maxChannelLetters = 5;
var maxTitleLetters   = 12;
var tableMargin       = 14;
var isFlipped         = false;

// filename of cache described in CSV format
var cacheFilename = "/tmp/cachefile.csv";

var monitoringTimer  = null;
var updateInterval   = 60; // in seconds
var trendInterval    = 3; // how frequent the trend is acquired
var trendCount       = 0;
var currentTrend     = new Array();
var trendMode        = false;

// wait timer 
var waitTimer   = null;
var waitCounter = 0;
var waitTimerImages = new Array();

// command that are able to be canceled
var loaderCommand;
var trendCommand = null;

// activity monitor
var activityLevels     = 5;
var activityBoundaries = [ 0, 25, 50, 100, 200 ];
var activityBackgroundImages = [ "Images/Level0.png", "Images/Level1.png", 
                                 "Images/Level2.png", "Images/Level3.png", 
                                 "Images/Level4.png" ];
var activityBackgroundColors = [ "#000", "#300", "#600", "#a00", "#f00" ];

// debug output
function debug( message ) {
    if ( LOGGING == false ) return;
    if ( window.TVTableGenerator ) {
        TVTableGenerator.log( message );
    } else {
        var elem = document.getElementById( 'debugDiv' );
        if ( elem ) {
            if ( elem.style.display != "none" ) {
                elem.appendChild( document.createTextNode( message ) );
                elem.appendChild( document.createElement( 'br' ) );
            }
        }
    }
}


// initializing parameters
function setup( event ) {
    _readFontSizes();

    // read parameters  from HTML
    importParameters();
    // read parameters from preferences
    ProgramArea = readPreference( "area", "tokyo" );
    ProgramWave = readPreference( "wave", "vhf" );
    trendMode   = ( readPreference( "trend", '' + trendMode ) == "true" );
    setDefaultSelections();

    // get parameters from HTML
    autoCollapse = ( getParameter( 'autoCollapse', 'true' ) != 'false' );
    updateInterval   = getParameter( 'updateInterval', 60 );
    trendInterval    = getParameter( 'activityCheckInterval', 3 );
    if ( updateInterval < 60 ) updateInterval = 60;
    if ( trendInterval < 1 ) trendInterval = 1;
    {
        var dateline_str = getParameter( 'dateLine', '04:00' );
        var regex = new RegExp( "(\\d+):(\\d+)" );
        var items = regex.exec( dateline_str );
        if ( items && items.length == 3 ) {
            dateLine = parseInt( items[ 1 ] ) * 60 + parseInt( items[ 2 ] );
        }
        if ( dateLine > 240 ) dateLine = 240;
    }
    loadActivityParameters();

    // layout
    var front = document.getElementById( 'front' );
    // fit window size (window size is not correctly given after reloading with cmd+R)
    {
        var fs = document.defaultView.getComputedStyle( front );
        var fw = parseInt( fs.getPropertyValue( 'width' ) );
        var fh = parseInt( fs.getPropertyValue( 'height' ) );
        var fl = parseInt( fs.getPropertyValue( 'left' ) );
        var ft = parseInt( fs.getPropertyValue( 'top' ) );
        window.resizeTo( fl + fw + fl, fh + ft + ft );
    }

    expander = new Expander( front, 700, 200 );
    expander.maxSteps = parseInt( getParameter( 'collapseSteps', 10 ) );
    if ( expander.maxSteps < 2 ) expander.maxSteps = 2;
    expander.animationInterval = parseInt( getParameter( 'collapseWait', 33 ) );
    displaySpan = parseInt( getParameter( 'displaySpan', 180 ) );
    channelFormat = getParameter( 'channelFormat', "<$program-start$><$starttime$><$title$><$program-end$>" );
    if ( getParameter( 'expandBothSide', 'true' ) == 'false' ) {
        expander.positionFixed = true;
    } else {
        expander.positionFixed = false;
    }

    // button
    var doneButton = document.getElementById( 'done' );
    if ( doneButton ) {
        createGenericButton( doneButton, "Done", flipToFront );
    }
    createInformationButton( document.getElementById( "front" ), -10, -10, collapseToFlip, true, 500 );

    __setExpandButtonImage();

    // preload images
    loadTimerImages();

    // set event handlers
    if ( window.widget ) {
        widget.onremove = cleanUp;
        window.onblur = function() { 
            if ( autoCollapse ) {
                if ( expander && expander.isCollapsed == false ) {
                    toggleMode( null );
                }
            }
        };

        cacheFilename = "/tmp/television_guide_" + widget.identifier + ".csv";
        widget.onshow = startMonitor;
        widget.onhide = stopMonitor;
    }

    window.addEventListener( 'mousewheel', mouseWheel, true );
    window.addEventListener( 'keypress', keyPressed, true );

    // plugin
    try {
        if ( window.TVTableGenerator ) {
            debug( "plugin detected, loading " + cacheFilename );
            TVTableGenerator.setProgramFile( cacheFilename );
        } else {
            debug( "no plugin" );
        }
    } catch ( e ) {
        debug( e );
    }

    setButtonEvents();

    setTimeout( reload, 0 );

    // set monitor
    monitoringTimer = null;
    startMonitor();
}

function loadActivityParameters() {
    var levels = parseInt( getParameter( 'activityLevels', 5 ) );
    activityLevels = levels;

    var thresholds = getParameter( 'activityThresholds', '0,25,50,100,200' ).split( ',' );
    activityBoundaries = new Array();
    for ( var i = 0; i < levels; i++ ) {
        if ( i >= thresholds.length ) {
            activityBoundaries.push( 200 );
        } else {
            activityBoundaries.push( parseInt( thresholds[ i ] ) );
        }
    }

    var colors = getParameter( 'activeChannelColors', '#000,#300,#600,#a00,#f00' ).split( ',' );
    activityBackgroundColors = new Array();
    for ( var i = 0; i < levels; i++ ) {
        if ( i >= colors.length ) {
            activityBackgroundColors.push( 'blue' );
        } else {
            activityBackgroundColors.push( colors[ i ] );
        }
    }

    var backgrounds = getParameter( 'activeChannelBackgrounds', "Images/Level0.png,Images/Level1.png,Images/Level2.png,Images/Level3.png,Images/Level4.png" ).split( ',' );
    activityBackgroundImages = new Array();
    for ( var i = 0; i < levels; i++ ) {
        if ( i >= backgrounds.length ) {
            activityBackgroundImages.push( 'none' );
        } else {
            activityBackgroundImages.push( backgrounds[ i ] );
        }
    }

}

// preload images
function loadTimerImages() {
    waitTimerImages = new Array();
    for ( var i = 0; i < 12; i++ ) {
        var im = new Image();
        im.src = "Images/Wait" + i + ".png";
        waitTimerImages.push( "url(" + im.src + ")" );
    }
}

// import parameters from HTML
function importParameters() {
    importedParameters = new Array();
    var params = document.getElementsByTagName( 'parameter' );
    //debug( "PARAMS " + params.length );
    for ( var i = 0; i < params.length; i++ ) {
        try {
            var name = params[ i ].getAttribute( 'name' );
            var value = params[ i ].getAttribute( 'value' );
            //debug( "SET " + name + " : " + value );
            importedParameters[ name ] = value;
        } catch ( e ) {
            debug( e );
        }
    }
}

// get parameter value
function getParameter( label, defaultValue ) {
    if ( importedParameters == null ) {
        importParameters();
    }
    var params = document.getElementsByTagName( 'parameter' );
    if ( importedParameters[ label ] ) {
        return importedParameters[ label ];
    }
    return defaultValue;
}


// read preferences and set default values of preferences
function setDefaultSelections() {
    // area
    var elem_area = document.getElementById( 'area' );
    {
        var children = elem_area.childNodes;
        for ( var i = 0; i < children.length; i++ ) {
            var option = children.item( i );
            try {
                if ( option.value == ProgramArea ) {
                    option.setAttribute( 'selected' );
                } else {
                    option.removeAttribute( 'selected' );
                }
            } catch ( e ) {}
        }
    }

    // wave (VHF, UHF, BS)
    var elem_wave = document.getElementById( 'wave' );
    {
        var children = elem_wave.childNodes;
        for ( var i = 0; i < children.length; i++ ) {
            var option = children.item( i );
            try {
                if ( option.value == ProgramWave ) {
                    option.setAttribute( 'selected' );
                } else {
                    option.removeAttribute( 'selected' );
                }
            } catch ( e ) {}
        }
    }

    // whether trends are checked
    var elem_trend = document.getElementById( 'trend' );
    if ( trendMode ) {
        elem_trend.setAttribute( 'checked' );
    } else {
        elem_trend.removeAttribute( 'checked' );
    }

    if ( ProgramWave != 'vhf' ) {
        elem_trend.setAttribute( 'disabled' );
        elem_trend.removeAttribute( 'checked' );
    } else {
        elem_trend.removeAttribute( 'disabled' );
    }

}


// event listeners
function __appendClicker( key, id, style1, style2, handler ) {
    var button = document.getElementById( id );
    buttonlisteners[ key ] = new ClickListener( button, handler, style1, style2 );
}

function setButtonEvents() {
    //var buttonlisteners = new Array();
    __appendClicker( 'expand', 'expandbutton', 'button', 'button_activated', toggleMode );
    __appendClicker( 'reload', 'reloadbutton', 'button', 'button_activated', reload );

    __appendClicker( 'up', 'upperArrow', 'upper_default', 'upper_activated', programTableBackward );
    __appendClicker( 'home', 'homeButton', 'homeButton_default', 'homeButton_activated', loadChannels );
    __appendClicker( 'down', 'lowerArrow', 'lower_default', 'lower_activated', programTableForward );
}

function changeWaveTemporary( event ) {
    var elem = event.srcElement;
    if ( elem.nodeType == 3 ) {
        elem = elem.parentElement;
    }
    if ( elem.nodeType != 1 ) {
        debug( "event not fired from element" );
    }
    var wave = elem.value;
    //debug( wave );
    if ( wave != ProgramWave ) {
        killLoadingProcess();
        ProgramWave = wave;
        reload();
    }

}

// !reload
// reload programs
function reload(event) {
    stopWaitTimer();
    try {
        // stop process if reloading
        if ( loaderCommand ) {
            debug( "stop" );
            killLoadingProcess();
            var button = document.getElementById( 'reloadbutton' );
            if ( button ) {
                button.innerText = "Reload";
            }
            return;
        }
    } catch ( e ) {
        debug( "??? " + e );
    }

    // bright wave button
    var elem_wave = document.getElementById( 'wavebutton' );
    if ( elem_wave ) {
        elem_wave.setAttribute( 'class', 'button_activated' );
    }

    // execute Python script
    var option_offset = "";
    
    {
        var date = new Date();
        var current_min = date.getHours() * 60 + date.getMinutes();
        if ( current_min < dateLine ) {
            option_offset = " -offset \"-1\"";
        }
    }

    var pycommand = "/usr/bin/python loadtv2csv.py -area " + ProgramArea + " -o " + cacheFilename + " -wave " + ProgramWave + option_offset;
    debug( pycommand );
     if ( window.widget ) {
         startWaitTimer();
         loaderCommand = widget.system( pycommand, reloadEndHandler );
         var button = document.getElementById( 'reloadbutton' );
         if ( button ) {
             button.innerText = "Stop";
         }
     }

}


// receive reloading results
function reloadEndHandler( command ) {
    //debug( "endHandler " + command );
    loaderCommand = null;
    stopWaitTimer();
    var button = document.getElementById( 'reloadbutton' );
    if ( button ) {
        button.innerText = "Reload";
    }

    // display error and exit when any errors are detected
    var error = command.errorString;
    if ( error ) {
        showDescription( error );
    } else {
        // refresh
        dayOfData = new Date();

        if ( window.TVTableGenerator ) {
            // resize window
            var num_channels_prec = TVTableGenerator.getChannelSize();
            TVTableGenerator.setProgramFile( cacheFilename );
            var num_channels_current = TVTableGenerator.getChannelSize();
            // adjust widget size
            var expander_tmp = null;
            if ( expander && expander.isCollapsed == false ) {
                var width_channel = parseInt( getParameter( 'channelWidth', 80 ) );
                var width = expander.size1.width + width_channel * num_channels_current + tableMargin;
                expander_tmp = new Expander( expander.element, width, expander.size2.height );
                expander_tmp.size1.width = expander.size2.width;
                expander_tmp.size1.height = expander.size2.height;
                expander_tmp.maxSteps = expander.maxSteps / 2;
            }


            if ( expander_tmp == null || expander_tmp.size2.width == expander.size2.width ) {
                loadChannels();
            } else {
                expander.size2.width = expander_tmp.size2.width;
                expander_tmp.callBack = loadChannels;
                expander_tmp.expand();
            }
        }
    }

    // bright wave button
    var elem_wave = document.getElementById( 'wavebutton' );
    if ( elem_wave ) {
        elem_wave.setAttribute( 'class', 'button' );
    }

    // change selected item of temporary wave selection
    var elem_select = document.getElementById( 'wave-temp' );
    if ( elem_select ) {
        elem_select.value = ProgramWave;
        var children = elem_select.childNodes;
        for ( var i = 0; i < children.length; i++ ) {
            var option = children.item( i );
            //debug( i + " : " + option.value + " <= " + ProgramWave );
            if ( option.value ) {
                if ( option.value == ProgramWave ) {
                    option.setAttribute( 'selected' );
                } else {
                    option.removeAttribute( 'selected' );
                }
            }
        }
    }
}

function stopWaitTimer() {
    debug( "stop timer" );
    if ( waitTimer == null ) return;
    clearInterval( waitTimer );
    waitTimer = null;
    var elem_wait = document.getElementById( "wait" );
    if ( elem_wait ) {
        elem_wait.style.display = "none";
    }
}

function startWaitTimer() {
    debug( "start timer" );
    if ( waitTimer ) return;
    waitCounter = 0;
    waitTimer = setInterval( __tickWaitTimer, 66 );
    var elem_wait = document.getElementById( "wait" );
    if ( elem_wait ) {
        elem_wait.style.display = "block";
    }
}

function __tickWaitTimer() {
    var elem_wait = document.getElementById( "wait" );
    //debug( "counter " + waitCounter );
    if ( elem_wait ) {
        waitCounter = ( waitCounter + 1 ) % waitTimerImages.length;
        elem_wait.style.backgroundImage = waitTimerImages[ waitCounter ];
    }
}


function readPreference( key, defaultValue ) {
    if ( window.widget ) {
        var val_unique = widget.preferenceForKey( widget.identifier + "-" + key );
        if ( val_unique ) return val_unique;
        var val = widget.preferenceForKey( key );
        if ( val ) return val;
    }
    return defaultValue;
}

function writePreference( key, value ) {
    if ( window.widget ) {
        widget.setPreferenceForKey( value, widget.identifier + "-" + key );
        widget.setPreferenceForKey( value, key );
    }
}


function cleanUp() {
    // cancel python script
    killLoadingProcess();
    // remove widget's unique preference
    if ( window.widget ) {
        widget.setPreferenceForKey( null, widget.identifier + "-area" );
        widget.setPreferenceForKey( null, widget.identifier + "-wave" );
    }
    // remove cache file
    if ( window.TVTableGenerator ) {
        window.TVTableGenerator.removeProgramFile( cacheFilename );
    }
}

// !updateProgram
//
function updateProgram() {
    // reload programs?
    {
        var date_current = new Date();
        //var days = date_current.getDate();
        // same day but over dateline
        if ( date_current.getDate() == dayOfData.getDate() 
             && date_current.getMonth() == date_current.getMonth() ) {
            if ( dayOfData.getHours() * 60 + dayOfData.getMinutes() < dateLine
                 && date_current.getHours() * 60 + date_current.getHours() >= dateLine ) {
                reload();
            }
        } else { // different day
            var yesterday = new Date( date_current.getTime() - 3600 * 1000 * 24 );
            if ( yesterday.getMonth() == dayOfData.getMonth()
                 && yesterday.getDate() == dayOfData.getDate() ) {
                if ( date_current.getHours() * 60 + date_current.getMinutes() >= dateLine ){
                    reload();
                }
            } else {
                reload();
            }
        }
    }

    if ( expander ) {
        if ( expander.isCollapsed ) {
            // refresh currently boardcasting programs for collapsed mode
            if ( waitTimer == null ) { // do not update while loading
                //debug( "update program " );
                setCurrentPrograms();
            } 
        } else {
            loadChannels();
        }
    }

    // check trend
    if ( waitTimer == null && trendMode ) {
        debug( "trend count = " + trendCount );
        if ( trendCount == 0 ) {
            setTimeout( getTrend, 250 );
        }
        trendCount = ( trendCount + 1 ) % trendInterval;
    }
}

function startMonitor() {
    updateProgram();
//     if ( trendMode && trendCommand == null ) {
//         setTimeout( getTrend, 100 );
//     }
    if ( monitoringTimer == null ) {
        monitoringTimer = setInterval( updateProgram, 1000 * updateInterval );
        trendCount = 1;
    }
}

function resumeMonitor() {
    startMonitor();
}

function stopMonitor() {
    if ( monitoringTimer ) {
        clearInterval( monitoringTimer );
        monitoringTimer = null;
    }
}


function programTableForward( event ) {
    if ( tableStartTime ) {
        loadChannels( tableStartTime + offsetMinutes );
    }
}

function programTableBackward( event ) {
    if ( tableStartTime ) {
        loadChannels( tableStartTime - offsetMinutes );
    }
}

function showAreaAndTime( minutes, area ) {
    if ( minutes < 0 ) minutes += 1440;
    if ( minutes >= 1440 ) minutes -= 1440;
    var hs = parseInt( minutes / 60 );
    var ms = parseInt( minutes % 60 );
    var timeString;
    if ( hs < 10 ) {
        hs = "0" + hs;
    } else {
        hs = '' + hs;
    }
    if ( ms < 10 ) {
        ms = "0" + ms;
    } else {
        ms = '' + ms;
    }
    timeString = hs + ":" + ms;

    if ( area == undefined ) {
        area = ProgramArea;
    }
    var area_name = area;
    if ( window.TVTableGenerator ) {
        area_name = TVTableGenerator.getArea();
    }

    var elem_state = document.getElementById( 'areabutton' );
    if ( elem_state ) {
        if ( area_name == "" ) {
            elem_state.innerHTML = timeString;
        } else {
            elem_state.innerHTML = area_name + "/" + timeString;
        }
        //debug( area_name + ", " + timeString );
    }

    var elem_wave = document.getElementById( 'wavebutton' );
    if ( elem_wave ) {
        var children = elem_wave.childNodes;
        for ( var i = 0; i < children.length; i++ ) {
            var node = children.item( i );
            if ( node.nodeType != 1 ) { // elements are retained
                elem_wave.removeChild( node );
            }
        }
        var text = "?";
        if ( ProgramWave == "vhf" ) text = "VHF";
        else if ( ProgramWave == "uhf" ) text = "UHF/BS";
        else if ( ProgramWave == "bs_digital" ) text = "BS Digital";
        else if ( ProgramWave == "am" ) text = "AM";
        else if ( ProgramWave == "fm" ) text = "FM";
        elem_wave.appendChild( document.createTextNode( text ) );
    }
}


// !loadChannels
//  Write larger program table.
//  this function calls setCurrentPrograms() if the widget is collapsed
//
function loadChannels( startMin ) {
    var timeString = "";
    if ( startMin == undefined ) {
        var d = new Date();
        startMin = d.getHours() * 60 + d.getMinutes();
    }
    if ( startMin < 0 ) startMin += 1440;
    // set global variable
    tableStartTime = startMin;

    // inhibit morphing while morphing
    while ( expander.isMoving ) {
        var startmin_reserved = startMin;
        setTimeout( function() { loadChannels( startmin_reserved ); }, 100 );
        return;
    }

    // display status
    showAreaAndTime( startMin );

    var size = TVTableGenerator.getChannelSize();
    var width_channel = parseInt( getParameter( 'channelWidth', 80 ) );
    var width = expander.size1.width + width_channel * size;
    if ( expander.isCollapsed ) {
        expander.size2.width  = width + tableMargin;
        expander.size2.height = parseInt( getParameter( 'expandHeight', expander.size2.height ) );
    }

    if ( expander != null && expander.isCollapsed ) {
        //debug( "current programs" );
        setCurrentPrograms();
    }

    var elem_table = document.getElementById( 'programTable' );

    // clean up programs
    var children = elem_table.childNodes;
    for ( var i = children.length - 1; i >= 0; i-- ) {
        var child = children.item( i );
        try {
            if ( child.getAttribute( 'class' ) == "channel" ) {
                elem_table.removeChild( child );
            }
        } catch ( e ) {
//            debug( "removing child failed : " + e );
        }
    }


    var startMin24 = startMin;
    if ( startMin < offsetThreshold ) {
        startMin24 += 1440;
    }

    var color_odd = getParameter( 'oddLaneColor', 'white' );
    var color_even = getParameter( 'evenLaneColor', 'white' );

    for ( var i = 0; i < size; i++ ) {
        if ( expander.isCollapsed == false ) {
            // layout overflow, too much programs required for the current widget size
            if ( i * width_channel + expander.size1.width 
                 + tableMargin >= expander.size2.width ) {
                break;
            }
        }

        // create element of a channel
        var elem = document.createElement( 'div' );
        var channelname = TVTableGenerator.getChannelName( i );
        // layout
        elem.style.left = ( width_channel * i ) + "px";
        elem.style.width = width_channel + "px";
        // attributes
        elem.setAttribute( 'class', 'channel' );
        elem.setAttribute( 'id', channelname );

        // name of the channel
        var name = document.createElement( 'span' );
        name.setAttribute( 'class', 'channelName' );
        name.setAttribute( 'name', channelname );
        name.appendChild( document.createTextNode( channelname ) );
        // register an event handler, open URL refering to trend
        name.addEventListener( 'click', openActiveURL, true );
        if ( trendMode ) { // emphasis in trend mode
            setActivatedChannel( name, channelname );
        }
        elem.appendChild( name );

        // programs
        var progs = document.createElement( 'div' );
        progs.setAttribute( 'class', 'programs' );
        if ( i % 2 == 0 ) {
            progs.style.backgroundColor = color_even;
        } else {
            progs.style.backgroundColor = color_odd;
        }
        progs.innerHTML = getProgramsWithFormat( i, channelFormat, startMin24, startMin24 + displaySpan );
        elem.appendChild( progs );

        // append programs into table
        elem_table.appendChild( elem );
    }
    resizeFont( 0 );
    //setTimeout( resizeFont, 0 );
}

function getProgramsWithFormat( channelnumber, format, start, end ) {
    if ( window.TVTableGenerator ) {
        //debug( "ch " + channelnumber + ", start " + start + ", end " + end );
        //getProgramListWithRange
        var pout = TVTableGenerator.getProgramListWithRange( format, channelnumber, start, end );
        //debug( pout );
        return pout;
        //return "skip plugin";
        //return "prog1<br/>prog2<br/>";
    } else {
        return "prog1<br/>prog2<br/>";
    }
}

function collapseToFlip( event ) {
    if ( expander.canActivate() == false ) return;
    if ( expander.isCollapsed ) {
        expansionReserved = false;
        flipToBack();
    } else {
        expansionReserved = true;
        expander.callBack = flipToBack;
        expander.collapse();
    }
}

// show preference panel
function flipToBack( event ) {
    debug( "flip to back" );
    isFlipped = true;
    changeToShortMode();
    expander.callBack = null;
    if ( window.widget ) {
        widget.prepareForTransition( "ToBack" );
    }
    var front = document.getElementById( 'front' );
    var behind = document.getElementById( 'behind' );
    var fstyle = document.defaultView.getComputedStyle( front );
    // adjust back position
    if ( expander ) {
        behind.style.top  = expander.location.y + "px";
        behind.style.left = expander.location.x + "px";
        behind.style.width  = expander.size1.width + "px";
        behind.style.height = expander.size1.height + "px";
    } else {
        behind.style.top  = fstyle.getPropertyValue( "top" ) + "px";
        behind.style.left = fstyle.getPropertyValue( "left" ) + "px";
        behind.style.width  = fstyle.getPropertyValue( "width" ) + "px";
        behind.style.height = fstyle.getPropertyValue( "height" ) + "px";
    }

    front.style.display  = "none";
    behind.style.display = "block";
    if ( window.widget ) {
        setTimeout( "widget.performTransition()", 0 );
    }
}

function flipToFront( event ) {
    if ( window.widget ) {
        widget.prepareForTransition( "ToFront" );
    }
    var front = document.getElementById( 'front' );
    var behind = document.getElementById( 'behind' );
    isFlipped = false;

    var reloadFlag = false;
    var areaElem = document.getElementById( 'area' );
    if ( areaElem ) {
        var areaName = areaElem.value;
        //debug( "selected area : " + areaName + " <= " + ProgramArea );
        if ( areaName != ProgramArea ) {
            ProgramArea = areaName;
            writePreference( "area", areaName );
            // reload
            reloadFlag = true;
        }
    }

    var waveElem = document.getElementById( 'wave' );
    if ( waveElem ) {
        var wave_name = waveElem.value;
        //debug( "selected wave : " + ProgramWave + " => " + wave_name );
        if ( ProgramWave != wave_name ) {
            ProgramWave = wave_name;
            writePreference( "wave", wave_name );
            // reload
            reloadFlag = true;
        }
    }

    // activity can be obtained only for VHF channels
    if ( ProgramWave == "vhf" ) {
        var trendElem = document.getElementById( 'trend' );
        if ( trendElem ) {
            //debug( "TREND : " + trendElem.checked );
            var flag = trendElem.checked;
            if ( flag != trendMode ) {
                writePreference( "trend", '' + flag );
                trendMode = flag; 
            }
        }
    }

    if ( reloadFlag ) {
        killLoadingProcess();
        setTimeout( reload, 0 );
        if ( trendMode ) {
            setTimeout( getTrend, 500 );
        }
    }

    // show front panel
    front.style.display = "block";
    behind.style.display = "none";
    if ( window.widget ) {
        if ( expansionReserved ) {
            setTimeout( "widget.performTransition();setTimeout( function(){ eval('toggleMode();'); }, 1000 );", 0 );
        } else {
            setTimeout( "widget.performTransition()", 0 );
        }
    }
}

// kill script
function killLoadingProcess() {
    if ( loaderCommand ) { 
        loaderCommand.cancel();
        loaderCommand = null;
    }
}

// !toggleMode
//
function toggleMode( event ) {
    if ( expander.isCollapsed ) {
        if ( window.TVTableGenerator ) {
            if ( TVTableGenerator.getChannelSize() <= 0 ) {
                return;
            }
        } else {
            return;
        }

        if ( expander ) {
            debug( "set expanding size " );
            var size = TVTableGenerator.getChannelSize();
            var width_channel = getParameter( 'channelWidth', 80 );
            var width = expander.size1.width + width_channel * size;
            expander.size2.width  = width + tableMargin;
            expander.size2.height = parseInt( getParameter( 'expandHeight', expander.size2.height ) );
            //debug( "size is ( " +expander.size2.width + ", " +  expander.size2.height + " )" );
            // change icon
            __setExpandButtonImage();

            loadChannels();
            expander.callBack = changeToTableMode;
        }
    } else {
        var tables = document.getElementById( "programTable" );
        tables.style.display = "none";

        expander.callBack = changeToShortMode;
    }
    if ( expander ) {
        expander.toggle(event);
    }
}

// expand
function changeToTableMode() {
    expander.callBack = null;
    var description = document.getElementById( 'description' );
    description.innerHTML = "";

    var tables = document.getElementById( "programTable" );
    tables.style.display = "block";

    // change icon
    __setExpandButtonImage();
}

// collapse
function changeToShortMode() {
    expander.callBack = null;
    var description = document.getElementById( 'description' );
    setCurrentPrograms();

    // change icon
    __setExpandButtonImage();
}

// change icon
function __setExpandButtonImage() {
    var button = document.getElementById( "expandIcon" );
    if ( button ) {
        if ( expander && expander.isCollapsed ) {
            if ( button.expandSrc ) {
                button.src = button.expandSrc;
            } else {
                button.src = "Images/Expand.png";
            }
        } else {
            if ( button.collapseSrc ) {
                button.src = button.collapseSrc;
            } else {
                button.src = "Images/Collapse.png";
            }
        }
    }
}

function __truncateStringWithLength( str, length_limit ) {
    var pos = 0;
    var len_sgl = 0;
    var len_dbl = 0;
        
    for ( var pos = 0; pos < str.length; pos ++ ) {
        var c = str.charAt( pos );
        if ( c >= ' ' && c <= 'z' ) {
            len_sgl ++;
        } else {
            len_dbl ++;
        }
        if ( len_dbl + len_sgl / 2 > length_limit ) {
            return str.substring( 0, pos );
        }
    }
    return str;
}

// !setCurrentPrograms
//
//
function setCurrentPrograms() {
    var contents = "";
    if ( window.TVTableGenerator ) {
        var date = new Date();
        var minutes = date.getHours() * 60 + date.getMinutes();
        if ( minutes < offsetThreshold ) {
            minutes += 1440;
        }
        contents = TVTableGenerator.getCurrentPrograms( minutes );
        showAreaAndTime( minutes );
        //debug( "set current programs : " + minutes );
    } else {
        return;
    }

    var rows = contents.split( "\n" );
    var table = document.createElement( "table" );
    for ( var i = 0; i < rows.length; i++ ) {
        var columns = rows[ i ].split( ',' );
        if ( columns.length < 2 ) continue;
        var row = document.createElement( 'tr' );
        {
            var datum = document.createElement( 'td' );
            datum.setAttribute( 'class', 'cncolumn' );

            var chname = columns[ 0 ];
            // channel name clickable 
            var name = document.createElement( 'span' );
            name.setAttribute( 'name', chname );
            name.addEventListener( 'click', openActiveURL, true );
            // color name for trend mode
            if ( trendMode ) {
                setActivatedChannel( name, chname );
            }

            chname = __truncateStringWithLength( chname, maxChannelLetters );
            name.appendChild( document.createTextNode( chname ) );

            datum.appendChild( name );

            row.appendChild( datum );
        }
        {
            var datum = document.createElement( 'td' );
            var title = __truncateStringWithLength( columns[ 1 ], maxTitleLetters );
            datum.appendChild( document.createTextNode( title ) );
            row.appendChild( datum );
        }
        table.appendChild( row );
    }

    var description = document.getElementById( 'description' );
    if ( description ) {
        description.innerHTML = "";
        description.appendChild( table );
    }
}

function showProgramDescription( event ) {

    var elem = event.srcElement;
    if ( elem.nodeType == 3 ) {
        elem = elem.parentElement;
    }
    if ( elem.nodeType != 1 ) {
        return;
    }
    if ( elem.tagName.toLowerCase() == 'span' ) {
        elem = elem.parentElement;
    }
    if ( event ) {
        if ( event.shiftKey ) return;
    }

    var html = elem.innerHTML;
    var desc = elem.description;
    var gcode = elem.gcode;
    var iepg = elem.iepg;
    html = html;
    if ( gcode ) {
        html += "<br/><span class=\"gcode\">" + gcode + "</span></div>";
        if ( iepg ) {
            html = html + "<span class=\"iepg\" onmouseup=\"widget.openURL('" + iepg + "');\">&nbsp;</span>";
        }
    }
    if ( desc ) {
        html += "<br/>" + desc;
    }
    showDescription( html );
}

function showDescription( text ) {
    var description = document.getElementById( 'description' );
    if ( description ) {
        description.innerHTML = text;
    }
}

// event listeners
// receive mouse wheel
function mouseWheel( event ) {
    if ( isFlipped ) return; // ignore events at preference mode
    var delta = event.wheelDelta;
    var clicker = null;
    if ( delta > 0 ) {
        clicker = buttonlisteners[ 'up' ];
    } else {
        clicker = buttonlisteners[ 'down' ];
    }
    if ( clicker ) clicker.click();
}

// receive key events
function keyPressed( event ) {
    var flag_shift = event.shiftKey;
    var flag_alt = event.altKey;
    var flag_ctrl = event.ctrlKey;
    var flag_meta = event.metaKey;
    var code = event.keyCode;
    debug( "keypressed " + event.keyCode );
    if ( isFlipped ) {
        if ( code == 3 || code == 13 ) {
            setTimeout( flipToFront, 0 );
        }
        return; // ignore events at preference mode
    }
    if ( !flag_shift && !flag_alt && !flag_ctrl && flag_meta && code == 105 ) {
        collapseToFlip();
    }
    //if ( flag_shift || flag_alt || flag_ctrl || flag_meta ) return;
    if ( flag_meta && flag_ctrl == false ) {
        if ( code == 63233 ) { // Cmd + 'Down'
            resizeFont( -1 );
        } else if ( code == 63232 ) { // Cmd + 'Up'
            resizeFont( 1 );
        }
    }
    if ( flag_alt || flag_ctrl || flag_meta ) return;
    if ( code == 63232 ) {// up
        var clicker = buttonlisteners[ 'up' ];
        if ( clicker ) clicker.click();
    } else if ( code == 63233 ) { //down 
        var clicker = buttonlisteners[ 'down' ];
        if ( clicker ) clicker.click();
    } else if ( code == 32 ) { // space
        if ( buttonlisteners[ 'expand' ] ) {
            buttonlisteners[ 'expand' ].click();
        }
    } else if ( code == 3 || code == 13 ) { // return
        if ( buttonlisteners[ 'home' ] ) {
            buttonlisteners[ 'home' ].click();
        }
    } else {
        return;
    }
    event.stopPropagation();
    event.preventDefault();
}

// read font size
function _readFontSizes() {
    try {
        var div = document.createElement( "div" );
        document.getElementsByTagName( "body" )[ 0 ].appendChild( div );
        var desc = document.getElementById( "description" );
        var chn =  document.createElement( "div" );
        chn.setAttribute( "class", "channel" );
        div.appendChild( chn );

        var st1 = document.defaultView.getComputedStyle( desc );
        var st2 = document.defaultView.getComputedStyle( chn );

        //desc.appendChild( document.createTextNode( "TEXT" ) );
        //debug( "font desc : " + st1.getPropertyValue( "font-size" ) );
        //debug( "font chn  : " + st2.getPropertyValue( "font-size" ) );
        var fs1 = st1.getPropertyValue( "font-size" ).toLowerCase();
        var fs2 = st2.getPropertyValue( "font-size" ).toLowerCase();
        var reg = new RegExp( "^(\\d+)([^\\d]+)$" );

        var m1 = reg.exec( fs1 );
        if ( m1 ) {
            fontSizeDesc = baseFontSizeDesc = parseInt( m1[ 0 ] );
            baseFontUnitDesc = m1[ 2 ];
        }
        var m2 = reg.exec( fs2 );
        if ( m2 ) {
            fontSizeChannel = baseFontSizeChannel = parseInt( m2[ 0 ] );
            baseFontUnitChannel = m2[ 2 ];
        }
        //debug( "M1 = " + m1 + ", M2 = " + m2 );
        
        div.setAttribute( "style", "display:none;" );
        document.getElementsByTagName( "body" )[ 0 ].removeChild( div );
    } catch ( e ) {
        debug( e );
    }
}

// resize font
function resizeFont( dif ) {
    if ( dif == null ) {
        dif = 0;
    } else {
        fontSizeDesc += dif;
        fontSizeChannel += dif;
        debug( "resizing font : " + dif );
    }
    if ( fontSizeDesc < baseFontSizeDesc / 2 ) {
        fontSizeDesc = baseFontSizeDesc / 2;
    }
    if ( fontSizeDesc > baseFontSizeDesc * 2 ) {
        fontSizeDesc = baseFontSizeDesc * 2;
    }
    if ( fontSizeChannel < baseFontSizeChannel / 2 ) {
        fontSizeChannel = baseFontSizeChannel / 2;
    }
    if ( fontSizeChannel > baseFontSizeChannel * 2 ) {
        fontSizeChannel = baseFontSizeChannel * 2;
    }
    //debug( fontSizeDesc + baseFontUnitDesc );
    //debug( fontSizeChannel + baseFontUnitChannel );
    
    document.getElementById( "description" ).style.fontSize = fontSizeDesc + baseFontUnitDesc;
    var divs = document.getElementsByTagName( "div" );
    for ( var i = 0; i < divs.length; i++ ) {
        if ( divs[ i ].getAttribute( "class" ) == "channel" ) {
            divs[ i ].style.fontSize = fontSizeChannel + baseFontUnitChannel;
        }
    }

    return;
}

// change wave
function changeWave( event ) {
    debug( 'change wave ' );
    var wave  = document.getElementById( 'wave' );
    var trend = document.getElementById( 'trend' );
    if ( wave && trend ) {
        if ( wave.value == 'vhf' ) {
            trend.removeAttribute( 'disabled' );
            if ( trendMode ) {
                trend.setAttribute( 'checked' );
            }
        } else {
            trend.setAttribute( 'disabled' );
            trend.removeAttribute( 'checked' );
        }
    }
}

// invoke trend reading script
function getTrend() {
    debug( "gettrend" );
    if ( window.widget ) {
        if ( ProgramWave == 'vhf' ) {
            if ( trendCommand ) {
                trendCommand.cancel();
                trendCommand = null;
            }
            var command = "/usr/bin/python getcurrenttrend.py";
            trendCommand = widget.system( command, trendHandler );
        }
    }
}

// process trend results
function trendHandler( command ) {
    debug( 'activity processing' );
    try {
        var outputs = command.outputString.split( "\n" );
        for ( var i = 0; i < outputs.length; i++ ) {
            //debug( outputs[ i ] );
            var items = outputs[ i ].split( ',' );
            if ( items.length >= 3 ) {
                var cname = items[ 0 ];
                var chits = parseInt( items[ 1 ] );
                var curl  = items[ 2 ];
                //debug( cname + " : " + chits + ", " + curl );
                currentTrend[ cname ] = { hit:chits, url:curl };
            }
        }

        if ( expander && expander.isCollapsed ) {
            setCurrentPrograms();
        } else { // change channel title
            emphasisChannelNames();
        }
    } catch ( e ) {
        debug( "error in trend acquisition : " + e );
        var error = command.errorString;
        if ( error ) {
            debug( error );
            // neglect error for activity monitor
            //showDescription( error );
        }
        //throw e;
    }
}


// emphasis labels along 2ch trenad activity
function setActivatedChannel( element, name ) {
    var obj = currentTrend[ name ];
    if ( obj ) {
        var hit = obj.hit;
        var url = obj.url;
        var level = 0;
        for ( var i = activityBoundaries.length - 1; i >= 0; i-- ) {
            if ( hit >= activityBoundaries[ i ] ) {
                level = i;
                break;
            }
        }
        if ( element.getAttribute( 'class' ) == 'channelName' ) { // large table
            var bgimg = activityBackgroundImages[ level ];
            element.style.backgroundImage = "url(" + bgimg + ")";
        } else {
            element.style.color = activityBackgroundColors[ level ];
        }
        element.url = url;
    } else {
        //debug( "activity is not prepared for " + name );
    }
}

// read activest thread of 2ch.net and open the URL
function openActiveURL( event ) {
    if ( event == null || event == undefined ) return;
    var elem = event.srcElement;
    if ( elem == undefined || elem == null ) return;

    // avoid Safari bug, event was fired from Text, not element
    if ( elem.nodeType == 3 ) { 
        elem = elem.parentElement;
    }
    
    if ( elem.nodeType != 1 ) {
        debug( "event fired from not an element : " + elem );
        return;
    }

    // channel name
    var chname = null;
    try {
        chname = elem.getAttribute( 'name' ); // first choice, name attribute
    } catch ( e ) {
        try {
            chname = elem.getAttribute( 'id' ); // second choice, id attribute
        } catch ( e ) {
            chname = elem.innerText;            // third choice, inner text
        }
    }

    debug( "open url for " + chname );
    if ( trendMode && chname ) {
        var trend = currentTrend[ chname ];
        if ( trend ) {
            startWaitTimer();
            debug( "resolve from " + trend.url );
            var pycommand = "/usr/bin/python getcurrenttrend.py " + trend.url;
            debug( pycommand );
            if ( window.widget ) {
                var output = widget.system( pycommand, null ).outputString;
                stopWaitTimer();
                if ( output ) {
                    var outputs = output.split( "\n" );
                    if ( outputs.length > 0 ) {
                        if ( output.length > 1 ) {
                            widget.openURL( outputs[ 1 ] );
                        } else {
                            widget.openURL( outputs[ 0 ] );
                        }
                    }
                }
            }
        } else {
            debug( "no data" );
        }
    }
}


// set background of channel name with activities
function emphasisChannelNames() {
    if ( trendMode == false ) return;

    // collapse
    var elem_desc = document.getElementById( 'description' );
    if ( elem_desc && expander && expander.isCollapsed ) {
        setCurrentPrograms();
    }

    // expand
    var elem_table = document.getElementById( 'programTable' );
    if ( elem_table == null ) return;
    var children = elem_table.childNodes;
    for ( var i = 0; i < children.length; i++ ) {
        try {
            var child = children.item( i );
            if ( child.nodeType == 1 ) { // element
                if ( child.getAttribute( 'class' ) == 'channel' ) {
                    var label = child.firstChild;
                    if ( label.getAttribute( 'class' ) == 'channelName' ) {
                        setActivatedChannel( label, child.id );
                    }
                }
            }
        } catch ( e ) {
            debug( "i = " + i + " " + child + " : " + child.nodeName + ", " + child.nodeType + ", " + child.nodeValue );
            throw e;
        }
    }
}
